import traceback
from typing import List

import astroid
from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker

from dslinter.util.exception_handler import ExceptionHandler
from dslinter.util.ast import AssignUtil


class PCAScalerChecker(BaseChecker):

    __implements__ = IAstroidChecker

    name = "pca-scaler"
    priority = -1
    msgs = {
        "W5508": (
            "Scaler is not used before PCA",
            "pca scaler checker",
            "To ensure a good result, use feature scaling before Principle Component Analysis."
        ),
    }
    options = ()

    PIPELINE = [
        "make_pipeline",
        "Pipeline",
    ]

    PCA = ["PCA","KernelPCA","SparsePCA","IncrementalPCA",]

    SCALER = ["RobustScaler", "StandardScaler", "MaxAbsScaler", "MinMaxScaler",]

    LEARNING_FUNCTIONS: List[str] = [
        "fit",
        "fit_transform",
        "transform",
    ]

    Variables = []


    def visit_call(self, node: astroid.Call):
        try:
            # If there is no scaler before a pca, rule is violated.
            if (
                node.func is not None
                and hasattr(node.func, "name")
                and node.func.name in self.PIPELINE
                and node.args is not None
            ):
                hasPCA = False
                hasScaler = False
                for arg in node.args:
                    if isinstance(arg, astroid.Call):
                        if PCAScalerChecker._call_initiates_scaler(arg):
                            hasScaler = True
                        if PCAScalerChecker._call_initiates_pca(arg):
                            hasPCA = True
                            break
                    if isinstance(arg, astroid.node_classes.List):
                        for kid in arg.get_children():
                            for item in kid.get_children():
                                if isinstance(item, astroid.Call):
                                    if PCAScalerChecker._call_initiates_scaler(item):
                                        hasScaler=True
                                    if PCAScalerChecker._call_initiates_pca(item):
                                        hasPCA=True
                            if hasPCA == True:
                                    break
                if hasPCA == True and hasScaler==False:
                    self.add_message("pca scaler checker", node=node)

            if (
                    node.func is not None
                    and hasattr(node.func, "attrname")
                    and node.func.attrname in self.LEARNING_FUNCTIONS
                    and self._expr_is_pca(node.func.expr)
                    and node.args is not None
            ):
                hasPCA = True
                hasScaler = False
                for arg in node.args:
                    if isinstance(arg, astroid.Name):
                        values = AssignUtil.assignment_values(arg)
                        for value in values:
                            if (
                                    isinstance(value, astroid.Call)
                                    and value.func is not None
                                    and hasattr(value.func, "attrname")
                                    and value.func.attrname in self.LEARNING_FUNCTIONS
                            ):
                                if(self._expr_is_scaler(value.func.expr)):
                                    hasScaler = True
                                elif(isinstance(value.func.expr, astroid.Call)
                                    and value.func.expr.func is not None
                                    and hasattr(value.func.expr.func, "attrname")
                                    and value.func.expr.func.attrname in self.LEARNING_FUNCTIONS
                                    and self._expr_is_scaler(value.func.expr.func.expr)
                                ):
                                    hasScaler=True
                if hasPCA == True and hasScaler==False:
                    self.add_message("pca scaler checker", node=node)


        except:  # pylint: disable=bare-except
            ExceptionHandler.handle(self, node)
            traceback.print_exc()

    @staticmethod
    def _call_initiates_pca(call: astroid.Call) -> bool:
        """
        Evaluate whether a Call node is initiating an pca.

        :param call: Call to evaluate.
        :return: True when an estimator is initiated.
        """
        return (
            call.func is not None
            and hasattr(call.func, "name")
            and call.func.name in PCAScalerChecker.PCA
        )

    @staticmethod
    def _call_initiates_scaler(call: astroid.Call) -> bool:
        """
        Evaluate whether a Call node is initiating an scaler.

        :param call: Call to evaluate.
        :return: True when an estimator is initiated.
        """
        return (
            call.func is not None
            and hasattr(call.func, "name")
            and call.func.name in PCAScalerChecker.SCALER
        )

    @staticmethod
    def _expr_is_pca(expr: astroid.node_classes.NodeNG) -> bool:
        """
        Evaluate whether the expression is an estimator.

        :param expr: Expression to evaluate.
        :return: True when the expression is an estimator.
        """
        if isinstance(expr, astroid.Call) and PCAScalerChecker._call_initiates_pca(expr):
            return True

        # If expr is a Name, check whether that name is assigned to an estimator.
        if isinstance(expr, astroid.Name):
            values = AssignUtil.assignment_values(expr)
            for value in values:
                if PCAScalerChecker._expr_is_pca(value):
                    return True
        return False

    @staticmethod
    def _expr_is_scaler(expr: astroid.node_classes.NodeNG) -> bool:
        """
        Evaluate whether the expression is an estimator.

        :param expr: Expression to evaluate.
        :return: True when the expression is an estimator.
        """
        if isinstance(expr, astroid.Call) and PCAScalerChecker._call_initiates_scaler(expr):
            return True

        # If expr is a Name, check whether that name is assigned to an estimator.
        if isinstance(expr, astroid.Name):
            values = AssignUtil.assignment_values(expr)
            for value in values:
                if PCAScalerChecker._expr_is_scaler(value):              
                    return True
        return False



"""Hyperparameter checker checks whether all hyperparameters for learning algorithms are set."""
from typing import Dict, List

import astroid
from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker

from dslinter.util.exception_handler import ExceptionHandler
from dslinter.util.resources import Resources


class HyperparameterChecker(BaseChecker):
    """Hyperparameter checker checks whether all hyperparameters for learning algorithms are set."""

    __implements__ = IAstroidChecker

    name = "hyperparameters"
    priority = -1
    msgs = {
        "W5505": (
            "Hyperparameter not set.",
            "hyperparameters",
            "For learning algorithms, hyperparameters should be tuned and set.",
        ),
    }
    options = (
        (
            "strict_hyperparameters",
            {
                "default": False,
                "type": "yn",
                "metavar": "<y_or_n>",
                "help": "Force that all parameters of learning algorithms are set.",
            },
        ),
    )

    # Main hyperparameters of learning algorithms, as defined in research.
    # Sources:
    # 1. Probst, P., Boulesteix, A. L., & Bischl, B. (2019). Tunability: Importance of
    #   Hyperparameters of Machine Learning Algorithms. Journal of Machine Learning Research,
    #   20(53), 1-32.
    # 2. van Rijn, J. N., & Hutter, F. (2018, July). Hyperparameter importance across datasets.
    #   In Proceedings of the 24th ACM SIGKDD International Conference on Knowledge Discovery &
    #   Data Mining (pp. 2367-2376).
    HYPERPARAMETERS_MAIN = {
        # sklearn.ensemble
        "AdaBoostClassifier": {"positional": 3, "keywords": ["learning_rate"]},
        "AdaBoostRegressor": {"positional": 3, "keywords": ["learning_rate"]},
        "GradientBoostingClassifier": {"positional": 2, "keywords": ["learning_rate"]},
        "GradientBoostingRegressor": {"positional": 2, "keywords": ["learning_rate"]},
        "HistGradientBoostingClassifier": {"positional": 2, "keywords": ["learning_rate"]},
        "HistGradientBoostingRegressor": {"positional": 2, "keywords": ["learning_rate"]},
        "RandomForestClassifier": {"positional": 7, "keywords": ["min_samples_leaf", "max_features"],},
        "RandomForestRegressor": {"positional": 7, "keywords": ["min_samples_leaf", "max_features"],},
        # sklearn.linear_model
        "ElasticNet": {"positional": 2, "keywords": ["alpha", "l1_ratio"]},
        # sklearn.neighbors
        "NearestNeighbors": {"positional": 1, "keywords": ["n_neighbors"]},
        # sklearn.svm
        "NuSVC": {"positional": 4, "keywords": ["nu", "kernel", "gamma"]},
        "NuSVR": {"positional": 4, "keywords": ["C", "kernel", "gamma"]},
        "SVC": {"positional": 4, "keywords": ["C", "kernel", "gamma"]},
        "SVR": {"positional": 4, "keywords": ["C", "kernel", "gamma"]},
        # sklearn.tree
        "DecisionTreeClassifier": {"positional": 14, "keywords": ["ccp_alpha"]},
        "DecisionTreeRegressor": {"positional": 14, "keywords": ["ccp_alpha"]},
    }

    def visit_call(self, node: astroid.Call):
        """
        When a Call node is visited, check whether hyperparameters are set.

        In strict mode, all hyperparameters should be set.
        In non-strict mode, function calls to learning functions should either contain all
        hyperparameters defined in HYPERPARAMETERS_MAIN or have at least one hyperparameter defined.

        :param node: Node which is visited.
        """
        try:
            try:
                function_name = node.func.name
            except AttributeError:
                return

            hyperparams_all = Resources.get_hyperparameters()

            if function_name in hyperparams_all:  # pylint: disable=unsupported-membership-test
                if self.config.strict_hyperparameters:
                    if not HyperparameterChecker._has_required_hyperparameters(node, hyperparams_all):
                        self.add_message("hyperparameters", node=node)
                else:  # non-strict
                    if (
                        function_name in self.HYPERPARAMETERS_MAIN
                        and not HyperparameterChecker._has_required_hyperparameters(node, self.HYPERPARAMETERS_MAIN)
                    ):
                        self.add_message("hyperparameters", node=node)
                    elif len(node.args) == 0 and node.keywords is None:
                        self.add_message("hyperparameters", node=node)
        except:  # pylint: disable=bare-except
            ExceptionHandler.handle(self, node)

    @staticmethod
    def _has_required_hyperparameters(node: astroid.Call, hyperparameters: Dict):
        """
        Evaluate whether a function call has all required hyperparameters defined.

        :param node: Node which is visited.
        :param hyperparameters: Dict of functions with their required hyperparameters.
        :return: True when all required hyperparameters are defined.
        """
        return len(node.args) >= hyperparameters[node.func.name]["positional"] or HyperparameterChecker._has_keywords(
            node.keywords, hyperparameters[node.func.name]["keywords"]
        )

    @staticmethod
    def _has_keywords(keywords: List[astroid.Keyword], keywords_goal: List[str]) -> bool:
        """
        Check if a list of keywords contains certain keywords.

        :param keywords: List of keywords.
        :param keywords_goal: Name of the keywords which are checked against.
        :return: True if keywords are present, False if they are not.
        """
        if keywords is None:
            return False

        found = 0
        for keyword in keywords:
            if keyword.arg in keywords_goal:
                found += 1

        return found == len(keywords_goal)



"""Checker which checks rules for excessive hyperparameter precision."""
import traceback
import astroid
from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker
from dslinter.util.exception_handler import ExceptionHandler
from dslinter.util.resources import Resources

import decimal

class ExcessiveHyperparameterPrecision(BaseChecker):
    """Checker which checks rules for excessive hyperparameter precision."""

    __implements__ = IAstroidChecker

    name = "excessive-hyperparameter-precision"
    priority = -1
    msgs = {
        "W5507": (
            "excessive hyperparameter precision might suggest over-tuning",
            "excessive hyperparameter precision",
            "excessive hyperparameter precision might suggest over-tuning"
        ),
    }
    options = ()

    highPrecisionCombinations={
        "KMeans":["tol"],
        "GraphicalLasso":[" tol","enet_tol"],
        "CCA":["tol"],"PLSCanonical":["tol"],
        "PLSRegression":["tol"],
        "GraphicalLassoCV":["tol","enet_tol"],
        "DictionaryLearning":["tol"],
        "FastICA":["tol"],
        "NMF":["tol"],
        "SparsePCA":["tol"],
        "LinearDiscriminantAnalysis":["tol"],
        "QuadraticDiscriminantAnalysis":["tol"],
        "GradientBoostingClassifier":["tol"],
        "GradientBoostingRegressor":["tol"],
        "HistGradientBoostingRegressor":["tol"],
        "HistGradientBoostingClassifier":["tol"],
        "GenericUnivariateSelect":["param"],
        "GaussianProcessRegressor": ["alpha"],
        "LogisticRegression":["tol"],
        "LogisticRegressionCV":["tol"],
        "Perceptron":["alpha"],
        "SGDClassifier":["alpha"],
        "SGDRegressor":["alpha"],
        "ElasticNet":["tol"],
        "ElasticNetCV":["tol"],
        "Lars":["eps"],
        "LarsCV":["eps"],
        "Lasso":["tol"],
        "LassoCV":["tol"],
        "LassoLars":["eps"],
        "LassoLarsCV":["eps"],
        "LassoLarsIC":["eps"],
        "ARDRegression":["alpha_1","alpha_2","lambda_1","lambda_2"],
        "BayesianRidge":["alpha_1","alpha_2","lambda_1","lambda_2"],
        "MultiTaskElasticNet":["tol"],
        "MultiTaskElasticNetCV":["tol"],
        "MultiTaskLasso":["tol"],
        "MultiTaskLassoCV":["tol"],
        "HuberRegressor":["alpha","tol"],
        "LocallyLinearEmbedding":["tol","hessian_tol","modified_tol"],
        "TSNE":["min_grad_norm"],
        "BayesianGaussianMixture":["reg_covar"],
        "GaussianMixture":["reg_covar"],
        "GaussianNB":["var_smoothing"],
        "NeighborhoodComponentsAnalysis":["tol"],
        "MLPClassifier":["alpha","tol"],
        "MLPRegressor":["alpha","tol"],
        "LinearSVC":["tol"],
        "LinearSVR":["tol"],
        }

    precisionThreshold = 3

    def visit_call(self, node: astroid.Call):
        """
        When a Call node is visited, check whether it violated the rules in this checker.

        :param node: The node which is visited.
        """
        try:
            try:
                function_name = node.func.name
            except AttributeError:
                return

            hyperparams_all = Resources.get_hyperparameters()
            if function_name in hyperparams_all:
                if(node.keywords is not None):
                    for keyword in node.keywords:
                        if(
                            function_name in self.highPrecisionCombinations
                            and keyword.arg in self.highPrecisionCombinations[function_name]
                        ):
                            continue
                        if (
                                hasattr(keyword, "value")
                                and hasattr(keyword.value, "value")
                                and type(keyword.value.value) == float
                                and decimal.Decimal(keyword.value.as_string()).as_tuple().exponent < 0
                                and abs(decimal.Decimal(keyword.value.as_string()).as_tuple().exponent) > self.precisionThreshold
                        ):
                            self.add_message("excessive hyperparameter precision", node=node)
        except:
            ExceptionHandler.handle(self, node)
            traceback.print_exc()


"""Main module for the plugin."""
from dslinter.checkers.data_leakage import DataLeakageChecker
from dslinter.checkers.dataframe import DataFrameChecker
from dslinter.checkers.hyperparameters import HyperparameterChecker
from dslinter.checkers.imports import ImportChecker
from dslinter.checkers.nan import NanChecker
from dslinter.checkers.controlling_randomness import ControllingRandomness
from dslinter.checkers.excessive_hyperparameter_precision import ExcessiveHyperparameterPrecision
from dslinter.checkers.pca_scaler import PCAScalerChecker


def register(linter):
    """
    Register checkers of the plugin with the linter.

    :param linter: Linter to add the checkers to.
    """
    linter.register_checker(ImportChecker(linter))
    linter.register_checker(HyperparameterChecker(linter))
    linter.register_checker(NanChecker(linter))
    linter.register_checker(DataFrameChecker(linter))
    linter.register_checker(DataLeakageChecker(linter))
    linter.register_checker(ControllingRandomness(linter))
    linter.register_checker(ExcessiveHyperparameterPrecision(linter))
    linter.register_checker(PCAScalerChecker(linter))
    



"""DataFrame checker which checks correct handling of calls on DataFrames."""
from typing import Dict, List

import astroid
from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker

from dslinter.util.ast import AssignUtil
from dslinter.util.exception_handler import ExceptionHandler
from dslinter.util.type_inference import TypeInference


class DataFrameChecker(BaseChecker):
    """DataFrame checker which checks correct handling of calls on DataFrames."""

    __implements__ = IAstroidChecker

    name = "dataframe"
    priority = -1
    msgs = {
        "W5501": (
            "Result of operation on a DataFrame is not assigned.",
            "unassigned-dataframe",
            "Most operations on a DataFrame return a new DataFrame. These should be assigned to \
            a variable.",
        ),
        "W5502": (
            "Iterating through a DataFrame.",
            "dataframe-iteration",
            "Iteration through a DataFrame is generally slow and should be avoided.",
        ),
        "W5503": (
            "Iterated object is modified.",
            "dataframe-iteration-modification",
            "An object where is iterated over should not be modified.",
        ),
    }
    options = ()

    _call_types: Dict[astroid.Call, str] = {}  # [node, inferred type of object the function is called on]

    # Whitelisted functions for which a DataFrame does not have to be assigned.
    WHITELISTED = [
        "all",
        "any",
        "bool",
        "boxplot",
        "corr",
        "corrwith",
        "count",
        "cov",
        "describe",
        "duplicated",
        "equals",
        "first_valid_index",
        "head",
        "hist",
        "idxmax",
        "idxmin",
        "info",
        "insert",
        "items",
        "iteritems",
        "iterrows",
        "itertuples",
        "keys",
        "last_valid_index",
        "mad",
        "max",
        "mean",
        "median",
        "memory_usage",
        "min",
        "mode",
        "nlargest",
        "nsmallest",
        "nunique",
        "plot",
        "profile_report",  # Jupyter notebook specific
        "sample",
        "sem",
        "skew",
        "sum",
        "tail",
        "update",
        "var",
        "to_***",
    ]

    def visit_module(self, node: astroid.Module):
        """
        When an Module node is visited, scan for Call nodes and get type the function is called on.

        :param node: Node which is visited.
        """
        try:
            # noinspection PyTypeChecker
            self._call_types = TypeInference.infer_types(node, astroid.Call, lambda x: x.func.expr.name)
        except:  # pylint: disable=bare-except
            ExceptionHandler.handle(self, node)

    def visit_call(self, node: astroid.Call):
        """
        When a Call node is visited, add messages if it violated the defined rules.

        :param node: Node which is visited.
        """
        try:
            if (
                self._is_simple_call_node(node)
                and not self._function_whitelisted(node)
                and self._dataframe_is_lost(node)
            ):
                self.add_message("unassigned-dataframe", node=node)
            if self._iterating_through_dataframe(node):
                self.add_message("dataframe-iteration", node=node)
        except:  # pylint: disable=bare-except
            ExceptionHandler.handle(self, node)

    @staticmethod
    def _is_simple_call_node(node: astroid.Call) -> bool:
        """
        Evaluate whether the node is a 'simple' call node.

        A 'simple' Call node is a single function call made on an expression.
        E.g., 'a.f()' and not 'f()' or 'a.f().g()' or 'a.f(g())'.

        :param node: Call node to evaluate.
        :return: True when the Call node is considered simple.
        """
        return (
            hasattr(node.func, "expr")  # The call is made on an expression.
            and hasattr(node.func.expr, "name")  # The expr the func is called on is a named thing.
            and not isinstance(node.parent, astroid.Attribute)  # Call is not an attribute.
            and not isinstance(node.parent, astroid.Call)  # Call is not part of another call.
        )

    @staticmethod
    def _function_whitelisted(node: astroid.Call) -> bool:
        """
        Evaluate whether the a Call node is whitelisted for the unassigned-dataframe rule.

        All whitelisted functions do not have to assigned to a variable. Whitelisted functions
        are listed in the WHITELISTED variable and also include all functions starting with "to_".

        :param node: Call node to evaluate.
        :return: True when the function of the call is whitelisted.
        """
        return hasattr(node.func, "attrname") and (
            node.func.attrname in DataFrameChecker.WHITELISTED or node.func.attrname[:3] == "to_"
        )

    def _dataframe_is_lost(self, node: astroid.Call) -> bool:
        """
        Check whether the call is done on a DataFrame and the result is lost.

        A result is seen as lost if its parent is an Expression and the operation is not done
        inplace.

        :param node: Node which is visited.
        :return: True when the call results in a DataFrame which is lost.
        """
        return (
            node in self._call_types  # Check if the type is inferred of this call.
            and (
                self._call_types[node] == "pandas.core.frame.DataFrame"
                or self._call_types[node] == "pyspark.sql.dataframe.DataFrame"
            )
            and not self._is_inplace_operation(node)
            # If the parent of the Call is an Expression, it means the DataFrame is lost.
            and isinstance(node.parent, astroid.Expr)
        )

    @staticmethod
    def _is_inplace_operation(node: astroid.Call) -> bool:
        """
        Evaluate whether the call has an 'inplace==True' keyword argument.

        :param node: Node to check the arguments from.
        :return: True when the call has an 'inplace==True' keyword argument.
        """
        if node.keywords is None:
            return False

        for keyword in node.keywords:
            if keyword.arg == "inplace":
                return keyword.value.value
        return False

    def _iterating_through_dataframe(self, node: astroid.Call) -> bool:
        """
        Evaluate whether there is iterated through a DataFrame.

        :param node: Node which is visited.
        :return: True when there is iterated through a DataFrame.
        """
        return (
            isinstance(node.parent, astroid.For)
            and node not in node.parent.body
            and node in self._call_types
            and (
                self._call_types[node] == "pandas.core.frame.DataFrame"
                or self._call_types[node] == "pyspark.sql.dataframe.DataFrame"
            )
        )

    def visit_for(self, node: astroid.For):
        """
        When a For node is visited, check for dataframe-iteration-modification violations.

        :param node: Node which is visited.
        """
        try:
            if not (
                isinstance(node.iter, astroid.Call)
                and node.iter in self._call_types
                and (
                    self._call_types[node.iter] == "pandas.core.frame.DataFrame"
                    or self._call_types[node.iter] == "pyspark.sql.dataframe.DataFrame"
                )
            ):
                return

            for_targets = DataFrameChecker._get_for_targets(node)
            assigned = AssignUtil.get_assigned_target_names(node)
            modified_iterated_targets = any(target in for_targets for target in assigned)

            if modified_iterated_targets:
                self.add_message("dataframe-iteration-modification", node=node)
        except:  # pylint: disable=bare-except
            ExceptionHandler.handle(self, node)

    @staticmethod
    def _get_for_targets(node: astroid.For) -> List[str]:
        """
        Get the target names of the for-loop definition.

        :param node: For node to get the target names from.
        :return: Target names.
        """
        target_names = []
        if isinstance(node.target, astroid.Tuple):
            for element in node.target.elts:
                if isinstance(element, astroid.AssignName):
                    target_names.append(element.name)
        elif isinstance(node.target, astroid.AssignName):
            target_names.append(node.target.name)
        return target_names



"""Setup script."""
from setuptools import find_packages, setup

# Get the long description from the README file
with open("README.md", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="dslinter",
    version="1.1.0",
    description="Pylint plugin for linting data science and machine learning code, focussed on the libraries pandas and scikit-learn.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Hynn01/dslinter",
    author="Mark Haakman, Haiyin Zhang, Daoyao Wang, Chadha Degachi",
    packages=find_packages(),
    package_data={"": ["*.pickle"]},
    python_requires="~=3.5",
    install_requires=["pylint~=2.0", "astroid~=2.4", "mypy", "data-science-types", "pyspark-stubs"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Plugins",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development :: Quality Assurance",
    ],
)



"""Checker which checks rules for controlling randomness."""
from typing import List
import traceback
import astroid
from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker
from dslinter.util.exception_handler import ExceptionHandler
from dslinter.util.resources import Resources


class ControllingRandomness(BaseChecker):
    """Checker which checks rules for controlling randomness."""

    __implements__ = IAstroidChecker

    name = "controlling-randomness"
    priority = -1
    msgs = {
        "W5506": (
            "'random_state=None' shouldn't be used in estimators or cross-validation splitters, it indicates improper randomness control",
            "controlling randomness",
            "For reproducible results across executions, remove any use of random_state=None."
        ),
    }
    options = ()

    SPLITTER_FUNCTIONS: List[str] = [
        "make_classification",
        "check_cv",
        "train_test_split",
    ]

    SPLITTER_CLASSES = [
        "GroupKFold",
        "GroupShuffleSplit",
        "KFold",
        "LeaveOneGroupOut",
        "LeavePGroupsOut",
        "LeaveOneOut",
        "LeavePOut",
        "PredefinedSplit",
        "RepeatedKFold",
        "RepeatedStratifiedKFold",
        "ShuffleSplit",
        "StratifiedKFold",
        "StratifiedShuffleSplit",
        "TimeSeriesSplit"
    ]

    def visit_call(self, node: astroid.Call):
        """
        When a Call node is visited, check whether it violated the rules in this checker.

        :param node: The node which is visited.
        """

        estimators_all = Resources.get_hyperparameters()
        try:
            if (
                hasattr(node, "func")
                and hasattr(node.func, "name")
                and hasattr(node, "keywords")
                and (node.func.name in self.SPLITTER_FUNCTIONS or node.func.name in self.SPLITTER_CLASSES or node.func.name in estimators_all)
                and node.keywords is not None
            ):
                for keyword in node.keywords:
                    if (keyword.arg == "random_state" and keyword.value.as_string() == "None"):
                        self.add_message("controlling randomness", node=node)
        except:
            ExceptionHandler.handle(self, node)
            traceback.print_exc()
